CLI_STATUS_STARTLINE = 5
XML_HEADERS = {
    "Accept-Encoding": "gzip",
    "X-Requested-With": "XMLHttpRequest",
    "User-Agent": "Go-http-client/1.1",
}
DOWNPOSTFIX = '.udown'
CACHEPOSTFIX = '.ucache'
DOWN_CHUNK_SIZE = 20480
OUTFILE_WRITE_BUF = 20480
DEFAULT_CONN_TIMEOUT = 30
MODEL_DOWNLOAD_URL = "https://github.com/JanPalasek/ulozto-captcha-breaker/releases/download/v2.2/model.tflite"
TOR_DATA_DIR_PREFIX = "tor_data_dir_"
MODEL_FILENAME = "model.tflite"
